#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from .. import client as _client
from .. import context as _context
from ..composition import (
    concurrency,
    type_inference,
)
from ..filters import _Filter
from ..identifiers import ProjectIdentifier
from ..retrieval import search

__all__ = ("list_containers",)


def list_containers(
    *,
    project_identifier: ProjectIdentifier,
    filter_: Optional[_Filter],
    context: Optional[_context.Context] = None,
    container_type: search.ContainerType,
) -> list[str]:
    validated_context = _context.validate_context(context or _context.get_context())
    client = _client.get_client(context=validated_context)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        )
        if inference_result.is_run_domain_empty():
            return []
        filter_ = inference_result.get_result_or_raise()

        sys_attr_pages = search.fetch_sys_id_labels(container_type)(client, project_identifier, filter_)
        return list(sorted(attrs.label for page in sys_attr_pages for attrs in page.items))



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from collections.abc import Collection
from typing import (
    Any,
    Generator,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd

from ..exceptions import ConflictingAttributeTypes
from . import identifiers
from .retrieval import (
    metrics,
    series,
)
from .retrieval.attribute_types import (
    TYPE_AGGREGATIONS,
    File,
)
from .retrieval.attribute_values import AttributeValue
from .retrieval.metrics import (
    IsPreviewIndex,
    PreviewCompletionIndex,
    StepIndex,
    TimestampIndex,
    ValueIndex,
)

__all__ = (
    "convert_table_to_dataframe",
    "create_metrics_dataframe",
    "create_series_dataframe",
    "create_files_dataframe",
)


def convert_table_to_dataframe(
    table_data: dict[str, list[AttributeValue]],
    selected_aggregations: dict[identifiers.AttributeDefinition, set[str]],
    type_suffix_in_column_names: bool,
    index_column_name: str = "experiment",
    flatten_aggregations: bool = False,
    flatten_file_properties: bool = False,
) -> pd.DataFrame:

    if flatten_aggregations:
        has_non_last_aggregations = any(aggregations != {"last"} for aggregations in selected_aggregations.values())
        if has_non_last_aggregations:
            raise ValueError("Cannot flatten aggregations when selected aggregations include more than just 'last'. ")

    if flatten_aggregations and flatten_file_properties:
        raise ValueError("Cannot set flatten_aggregations and flatten_file_properties at the same time")

    if not table_data and not flatten_aggregations:
        return pd.DataFrame(
            index=pd.Index([], name=index_column_name),
            columns=pd.MultiIndex.from_tuples([], names=["attribute", "aggregation"]),
        )
    if not table_data and flatten_aggregations:
        return pd.DataFrame(
            index=pd.Index([], name=index_column_name),
            columns=[],
        )

    def convert_row(values: list[AttributeValue]) -> dict[tuple[str, str], Any]:
        row = {}
        for value in values:
            column_name = get_column_name(value)
            if column_name in row:
                raise ConflictingAttributeTypes([value.attribute_definition.name])
            if value.attribute_definition.type in TYPE_AGGREGATIONS:
                aggregation_value = value.value
                selected_subset = selected_aggregations.get(value.attribute_definition, set())
                aggregations_set = TYPE_AGGREGATIONS[value.attribute_definition.type]

                agg_subset_values = get_aggregation_subset(aggregation_value, selected_subset, aggregations_set)

                for agg_name, agg_value in agg_subset_values.items():
                    row[(column_name, agg_name)] = agg_value
            elif flatten_file_properties and value.attribute_definition.type == "file":
                file_properties: File = value.value
                row[(column_name, "path")] = file_properties.path
                row[(column_name, "size_bytes")] = file_properties.size_bytes
                row[(column_name, "mime_type")] = file_properties.mime_type
            else:
                row[(column_name, "")] = value.value
        return row

    def flatten_row(row: dict[tuple[str, str], Any]) -> dict[str, Any]:
        """
        Flatten the row by converting tuple keys to string keys.
        """
        for (attribute, aggregation), value in row.items():
            if aggregation not in ("", "last"):
                raise ValueError(
                    f"Unexpected aggregation '{aggregation}' for attribute '{attribute}'. "
                    "Only 'last' or empty aggregation are allowed when flattening."
                )
        return {attribute: value for (attribute, aggregation), value in row.items()}

    def get_column_name(attr: AttributeValue) -> str:
        return f"{attr.attribute_definition.name}:{attr.attribute_definition.type}"

    def get_aggregation_subset(
        aggregations_value: Any, selected_subset: set[str], aggregations_set: Collection[str]
    ) -> dict[str, Any]:
        result = {}
        for agg_name in aggregations_set:
            if agg_name in selected_subset:
                result[agg_name] = getattr(aggregations_value, agg_name)
        return result

    def transform_column_names(df: pd.DataFrame) -> pd.DataFrame:
        if type_suffix_in_column_names:
            return df

        # Transform the column by removing the type
        original_columns = df.columns
        df.columns = [
            (col[0].rsplit(":", 1)[0], col[1]) if isinstance(col, tuple) else col.rsplit(":", 1)[0]
            for col in df.columns
        ]

        # Check for duplicate names
        duplicated_names = df.columns[df.columns.duplicated(keep=False)]  # type: ignore
        duplicated_names_set = set(duplicated_names)
        if duplicated_names.any():
            conflicting_types: dict[str, set[str]] = {}
            for original_col, new_col in zip(original_columns, df.columns):
                if isinstance(new_col, str):
                    continue

                if new_col in duplicated_names_set:
                    conflicting_types.setdefault(new_col[0], set()).add(original_col[0].rsplit(":", 1)[1])

            raise ConflictingAttributeTypes(conflicting_types.keys())  # TODO: add conflicting types to the exception

        return df

    rows: list[dict[Union[str, tuple[str, str]], Any]] = []
    for label, values in table_data.items():
        row: dict[Union[str, Union[str, tuple[str, str]]], Any] = convert_row(values)  # type: ignore
        if flatten_aggregations:
            # Note for future optimization:
            # flatten_aggregations is always True in v1
            # flatten_aggregations is always False in alpha
            row = flatten_row(row)
        row[index_column_name] = label
        rows.append(row)

    dataframe = pd.DataFrame(rows)
    dataframe = transform_column_names(dataframe)
    dataframe.set_index(index_column_name, drop=True, inplace=True)

    if not flatten_aggregations:
        dataframe.columns = pd.MultiIndex.from_tuples(dataframe.columns, names=["attribute", "aggregation"])

    sorted_columns = sorted(dataframe.columns)
    dataframe = dataframe[sorted_columns]

    return dataframe


def create_metrics_dataframe(
    metrics_data: dict[identifiers.RunAttributeDefinition, list[metrics.FloatPointValue]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
    *,
    type_suffix_in_column_names: bool,
    include_point_previews: bool,
    index_column_name: str,
    timestamp_column_name: Optional[str] = None,
) -> pd.DataFrame:
    """
    Creates a memory-efficient DataFrame directly from FloatPointValue tuples.

    Note that `data_points` must be sorted by (experiment name, path) to ensure correct
    categorical codes.

    There is an intermediate processing step where we represent paths as categorical codes
    Example:

    Assuming there are 2 user columns called "foo" and "bar", 2 steps each. The intermediate
    DF will have this shape:

            experiment  path   step  value
        0     exp-name     0    1.0    0.0
        1     exp-name     0    2.0    0.5
        1     exp-name     1    1.0    1.5
        1     exp-name     1    2.0    2.5

    `path_mapping` would contain {"foo": 0, "bar": 1}. The column names will be restored before returning the
    DF, which then can be sorted based on its columns.

    The reason for the intermediate representation is that logging a metric called eg. "step" would conflict
    with our "step" column during df.reset_index(), and we would crash.
    Operating on integer codes is safe, as they can never appear as valid metric names.

    If `timestamp_column_name` is provided, timestamp will be included in the DataFrame under the
    specified column.
    """

    path_mapping: dict[str, int] = {}
    sys_id_mapping: dict[str, int] = {}
    label_mapping: list[str] = []

    for run_attr_definition in metrics_data:
        if run_attr_definition.run_identifier.sys_id not in sys_id_mapping:
            sys_id_mapping[run_attr_definition.run_identifier.sys_id] = len(sys_id_mapping)
            label_mapping.append(sys_id_label_mapping[run_attr_definition.run_identifier.sys_id])

        if run_attr_definition.attribute_definition.name not in path_mapping:
            path_mapping[run_attr_definition.attribute_definition.name] = len(path_mapping)

    def generate_categorized_rows() -> Generator[Tuple, None, None]:
        for attribute, points in metrics_data.items():
            exp_category = sys_id_mapping[attribute.run_identifier.sys_id]
            path_category = path_mapping[attribute.attribute_definition.name]

            for point in points:
                # Only include columns that we know we need. Note that the list of columns must match the
                # the list of `types` below.
                head = (
                    exp_category,
                    path_category,
                    point[StepIndex],
                    point[ValueIndex],
                )
                if include_point_previews and timestamp_column_name:
                    tail: Tuple[Any, ...] = (
                        point[TimestampIndex],
                        point[IsPreviewIndex],
                        point[PreviewCompletionIndex],
                    )
                elif timestamp_column_name:
                    tail = (point[TimestampIndex],)
                elif include_point_previews:
                    tail = (point[IsPreviewIndex], point[PreviewCompletionIndex])
                else:
                    tail = ()

                yield head + tail

    types = [
        (index_column_name, "uint32"),
        ("path", "uint32"),
        ("step", "float64"),
        ("value", "float64"),
    ]

    if timestamp_column_name:
        types.append((timestamp_column_name, "uint64"))

    if include_point_previews:
        types.append(("is_preview", "bool"))
        types.append(("preview_completion", "float64"))

    df = pd.DataFrame(
        np.fromiter(generate_categorized_rows(), dtype=types),
    )

    experiment_dtype = pd.CategoricalDtype(categories=label_mapping)
    df[index_column_name] = pd.Categorical.from_codes(df[index_column_name], dtype=experiment_dtype)
    if timestamp_column_name:
        df[timestamp_column_name] = pd.to_datetime(df[timestamp_column_name], unit="ms", origin="unix", utc=True)

    df = _pivot_and_reindex_df(df, include_point_previews, index_column_name, timestamp_column_name)
    df = _restore_path_column_names(df, path_mapping, "float_series" if type_suffix_in_column_names else None)
    df = _sort_indices(df)

    return df


def create_series_dataframe(
    series_data: dict[identifiers.RunAttributeDefinition, list[series.SeriesValue]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
    index_column_name: str,
    timestamp_column_name: Optional[str],
) -> pd.DataFrame:
    experiment_mapping: dict[identifiers.SysId, int] = {}
    path_mapping: dict[str, int] = {}
    label_mapping: list[str] = []

    for run_attr_definition in series_data.keys():
        if run_attr_definition.run_identifier.sys_id not in experiment_mapping:
            experiment_mapping[run_attr_definition.run_identifier.sys_id] = len(experiment_mapping)
            label_mapping.append(sys_id_label_mapping[run_attr_definition.run_identifier.sys_id])

        if run_attr_definition.attribute_definition.name not in path_mapping:
            path_mapping[run_attr_definition.attribute_definition.name] = len(path_mapping)

    def generate_categorized_rows() -> Generator[Tuple, None, None]:
        for attribute, values in series_data.items():
            exp_category = experiment_mapping[attribute.run_identifier.sys_id]
            path_category = path_mapping[attribute.attribute_definition.name]

            if timestamp_column_name:
                for point in values:
                    yield exp_category, path_category, point.step, point.value, point.timestamp_millis
            else:
                for point in values:
                    yield exp_category, path_category, point.step, point.value

    types = [
        (index_column_name, "uint32"),
        ("path", "uint32"),
        ("step", "float64"),
        ("value", "object"),
    ]
    if timestamp_column_name:
        types.append((timestamp_column_name, "uint64"))

    df = pd.DataFrame(
        np.fromiter(generate_categorized_rows(), dtype=types),
    )

    experiment_dtype = pd.CategoricalDtype(categories=label_mapping)
    df[index_column_name] = pd.Categorical.from_codes(df[index_column_name], dtype=experiment_dtype)
    if timestamp_column_name:
        df[timestamp_column_name] = pd.to_datetime(df[timestamp_column_name], unit="ms", origin="unix", utc=True)

    df = _pivot_and_reindex_df(df, False, index_column_name, timestamp_column_name)
    df = _restore_path_column_names(df, path_mapping, None)
    df = _sort_indices(df)

    return df


def _pivot_and_reindex_df(
    df: pd.DataFrame,
    include_point_previews: bool,
    index_column_name: str,
    timestamp_column_name: Optional[str],
) -> pd.DataFrame:
    if df.empty and timestamp_column_name:
        # Handle empty DataFrame case to avoid pandas dtype errors
        df[timestamp_column_name] = pd.Series(dtype="datetime64[ns]")

    if include_point_previews or timestamp_column_name:
        # if there are multiple value columns, don't specify them and rely on pandas to create the column multi-index
        df = df.pivot(index=[index_column_name, "step"], columns="path")
    else:
        # when there's only "value", define values explicitly, to make pandas generate a flat index
        df = df.pivot(index=[index_column_name, "step"], columns="path", values="value")

    df = df.reset_index()
    df[index_column_name] = df[index_column_name].astype(str)
    df = df.sort_values(by=[index_column_name, "step"], ignore_index=True)
    df = df.set_index([index_column_name, "step"])

    return df


def _restore_path_column_names(
    df: pd.DataFrame, path_mapping: dict[str, int], type_suffix: Optional[str]
) -> pd.DataFrame:
    """
    Accepts an DF in an intermediate format in _create_dataframe, and the mapping of column names.
    Restores colum names in the DF based on the mapping.
    """

    # We need to reverse the mapping to index -> column name
    if type_suffix:
        reverse_mapping = {index: f"{path}:{type_suffix}" for path, index in path_mapping.items()}
    else:
        reverse_mapping = {index: path for path, index in path_mapping.items()}
    return df.rename(columns=reverse_mapping)


def _sort_indices(df: pd.DataFrame) -> pd.DataFrame:
    # MultiIndex DFs need to have column index order swapped: value/metric_name -> metric_name/value.
    # We also sort columns, but only after the original names have been restored.
    if isinstance(df.columns, pd.MultiIndex):
        df.columns.names = (None, None)
        df = df.swaplevel(axis=1)
        return df.sort_index(axis=1, level=0)
    else:
        df.columns.name = None
        return df.sort_index(axis=1)


def create_files_dataframe(
    files_data: list[tuple[identifiers.RunIdentifier, identifiers.AttributeDefinition, Optional[pathlib.Path]]],
    sys_id_label_mapping: dict[identifiers.SysId, str],
    index_column_name: str = "experiment",
) -> pd.DataFrame:
    if not files_data:
        return pd.DataFrame(
            index=pd.Index([], name=index_column_name),
        )

    rows: list[dict[str, Optional[str]]] = []
    for run_identifier, attribute_definition, target_path in files_data:
        row = {
            index_column_name: sys_id_label_mapping[run_identifier.sys_id],
            "attribute": attribute_definition.name,
            "file_path": str(target_path) if target_path else None,
        }
        rows.append(row)

    dataframe = pd.DataFrame(rows)
    dataframe = dataframe.pivot(index=[index_column_name], columns="attribute", values="file_path")

    sorted_columns = sorted(dataframe.columns)
    return dataframe[sorted_columns]



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pathlib
from typing import (
    Generator,
    Optional,
)

import pandas as pd

from .. import client as _client
from .. import (
    identifiers,
    output_format,
)
from ..composition import attribute_components as _components
from ..composition import (
    concurrency,
    type_inference,
    validation,
)
from ..context import (
    Context,
    get_context,
    validate_context,
)
from ..filters import (
    _AttributeFilter,
    _Filter,
)
from ..retrieval import (
    files,
    search,
)
from ..retrieval.search import ContainerType


def download_files(
    *,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[_Filter],
    attributes: _AttributeFilter,
    destination: pathlib.Path,
    context: Optional[Context],
    container_type: ContainerType,
) -> pd.DataFrame:
    valid_context = validate_context(context or get_context())
    client = _client.get_client(context=valid_context)

    attributes = validation.restrict_attribute_filter_type(attributes, type_in={"file"})
    validation.ensure_write_access(destination)

    with (
        concurrency.create_thread_pool_executor() as executor,
        concurrency.create_thread_pool_executor() as fetch_attribute_definitions_executor,
    ):
        inference_result = type_inference.infer_attribute_types_in_filter(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            executor=executor,
            fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
            container_type=container_type,
        )
        if inference_result.is_run_domain_empty():
            return output_format.create_files_dataframe(
                [],
                {},
                index_column_name="experiment" if container_type == search.ContainerType.EXPERIMENT else "run",
            )
        filter_ = inference_result.get_result_or_raise()

        sys_id_label_mapping: dict[identifiers.SysId, str] = {}

        def go_fetch_sys_attrs() -> Generator[list[identifiers.SysId], None, None]:
            for page in search.fetch_sys_id_labels(container_type)(
                client=client,
                project_identifier=project_identifier,
                filter_=filter_,
            ):
                sys_ids = []
                for item in page.items:
                    sys_id_label_mapping[item.sys_id] = item.label
                    sys_ids.append(item.sys_id)
                yield sys_ids

        output = concurrency.generate_concurrently(
            items=go_fetch_sys_attrs(),
            executor=executor,
            downstream=lambda sys_ids: _components.fetch_attribute_definition_aggregations_split(
                client=client,
                project_identifier=project_identifier,
                attribute_filter=attributes,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids,
                downstream=lambda sys_ids_split, definitions_page, _: _components.fetch_attribute_values_split(
                    client=client,
                    project_identifier=project_identifier,
                    executor=executor,
                    sys_ids=sys_ids_split,
                    attribute_definitions=definitions_page.items,
                    downstream=lambda values_page: concurrency.generate_concurrently(
                        items=(
                            (value, file_)
                            for value, file_ in zip(
                                values_page.items,
                                files.fetch_signed_urls(
                                    client=client,
                                    project_identifier=project_identifier,
                                    file_paths=[value.value.path for value in values_page.items],
                                ),
                            )
                        ),
                        executor=executor,
                        downstream=lambda run_file_tuple: concurrency.return_value(
                            (
                                run_file_tuple[0].run_identifier,
                                run_file_tuple[0].attribute_definition,
                                files.download_file_retry(
                                    client=client,
                                    project_identifier=project_identifier,
                                    signed_file=run_file_tuple[1],
                                    target_path=files.create_target_path(
                                        destination=destination,
                                        experiment_name=sys_id_label_mapping[run_file_tuple[0].run_identifier.sys_id],
                                        attribute_path=run_file_tuple[0].attribute_definition.name,
                                    ),
                                ),
                            )
                        ),
                    ),
                ),
            ),
        )

        results: Generator[
            tuple[identifiers.RunIdentifier, identifiers.AttributeDefinition, Optional[pathlib.Path]],
            None,
            None,
        ] = concurrency.gather_results(output)
        file_list = list(results)

        return output_format.create_files_dataframe(
            file_list,
            sys_id_label_mapping,
            index_column_name="experiment" if container_type == search.ContainerType.EXPERIMENT else "run",
        )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import warnings
from typing import (
    Any,
    Dict,
    Iterable,
    Optional,
    Sequence,
)

from neptune_query.internal import env


class NeptuneError(Exception):
    def __init__(self, message: str, **kwargs: Any) -> None:
        super().__init__(message.format(**kwargs, **_styles))


class NeptuneWarning(Warning):
    def __init__(self, message: str, **kwargs: Any) -> None:
        super().__init__(message.format(**kwargs, **_styles))


class NeptuneUserError(NeptuneError):
    def __init__(self, message: str, **kwargs: Any) -> None:
        super().__init__(message, **kwargs)


class NeptuneProjectNotProvided(NeptuneUserError):
    def __init__(self) -> None:
        super().__init__(
            """
{h1}NeptuneProjectNotProvided: The project name was not provided.{end}

Make sure to specify a valid project in one of the following ways:

- Call the `set_project()` function
- Create a Context with the project name and pass it to the `context` argument of the fetching method
- Set the `NEPTUNE_PROJECT` environment variable

For details, see https://docs.neptune.ai/fetcher_setup
"""
        )


class NeptuneFailedToFetchClientConfig(NeptuneError):
    def __init__(self, exception: Exception) -> None:
        super().__init__(
            """
{h1}NeptuneFailedToFetchClientConfigException: Failed to fetch the client configuration.{end}

This can happen due to network issues or if Neptune is unreachable.

Make sure that you have a stable internet connection and that your Neptune instance is running and accessible.

If the issue persists, please contact Neptune support with the following information:

{exception}
""",
            exception=exception,
        )


class NeptuneProjectInaccessible(NeptuneError):
    def __init__(self) -> None:
        super().__init__(
            """
{h1}NeptuneProjectInaccessible: You don't have access to the project or it doesn't exist.{end}

Ensure that:
- the workspace and project names are correct
- the account you're using has at least Viewer access to the project

For details, see https://docs.neptune.ai/project_access
"""
        )


class NeptuneApiTokenNotProvided(NeptuneUserError):
    def __init__(self) -> None:
        super().__init__(
            """
{h1}NeptuneApiTokenNotProvided: The Neptune API token was not provided.{end}

Make sure to specify a valid token in one of the following ways:

- Call the `set_api_token()` function
- Create a Context with the API token and pass it to the `context` argument of the fetching method
- Set the `NEPTUNE_API_TOKEN` environment variable

For details, see https://docs.neptune.ai/fetcher_setup
"""
        )


class NeptuneInvalidCredentialsError(NeptuneUserError):
    def __init__(self) -> None:
        super().__init__(
            """
{h1}NeptuneInvalidCredentialsError: Your Neptune API token was rejected by the server.{end}

The API token must be valid and the associated account must have access to the target project.
Make sure to specify your Neptune credentials in one of the following ways:

- Call the `set_api_token()` or `set_project()` functions
- Create a Context with the API token and project, then pass it to the `context` argument of the fetching method
- Set the `NEPTUNE_API_TOKEN` and `NEPTUNE_PROJECT` environment variables

For details, see:
https://docs.neptune.ai/fetcher_setup
https://docs.neptune.ai/project_access
"""
        )


class AttributeTypeInferenceError(NeptuneError):
    def __init__(self, attribute_names: Sequence[str], details: Optional[Iterable[str]] = None) -> None:
        attribute_str = (
            f"attribute {attribute_names[0]}"
            if len(attribute_names) == 1
            else f"attributes [{', '.join(attribute_names)}]"
        )

        super().__init__(
            """
{h1}AttributeTypeInferenceError: Failed to infer types for {attribute_names}{end}
{details}
To resolve this ambiguity, specify the attribute type explicitly when constructing a filter:
    {python}
    fetch_experiments_table(
        experiments=Filter.eq(Attribute("metrics/m1", aggregation="max", type="float_series"), 1.2),
        ...
        sort_by=Attribute("config/batch_size", type="int"),
        ...
    )

    fetch_metrics(
        experiments=Filter.eq(Attribute("config/batch_size", type="int"), 64)
                    | Filter.eq(Attribute("config/batch_size", type="float"), 64.0),
        ...
    )

    list_attributes(
        experiments=Filter.eq(Attribute("config/batch_size", type="int"), 64),
        ...
    )
    {end}

For details, see https://docs.neptune.ai/attribute_types
""",
            attribute_names=attribute_str,
            details="\n" + "\n".join(details) if details else "",
        )


class ConflictingAttributeTypes(NeptuneError):
    def __init__(self, attribute_names: Iterable[str]) -> None:
        super().__init__(
            """
{h1}ConflictingAttributeTypes: Multiple types detected for attributes [{attribute_names}]{end}

Neptune found the attribute name in multiple runs, but the attribute type is not the same across all runs.

- To present each type as a separate column, use {python}type_suffix_in_column_names=True{end}.
- Alternatively, specify the attribute type explicitly using {python}AttributeFilter(..., type_in=[...]){end}.

For details, see https://docs.neptune.ai/attribute_types
""",
            attribute_names=", ".join(attribute_names),
        )


class NeptuneUnexpectedResponseError(NeptuneError):
    def __init__(self, status_code: int, content: bytes) -> None:
        super().__init__(
            """
{h1}NeptuneUnexpectedResponseError: The Neptune server returned an unexpected response.{end}

Response status: {status_code}
Response content: {content}
""",
            status_code=status_code,
            content=_decode_content(content),
        )


class NeptuneRetryError(NeptuneError):
    def __init__(
        self, retries: int, time: float, last_status_code: Optional[int] = None, last_content: Optional[bytes] = None
    ) -> None:
        content_str = _decode_content(last_content) if last_content else ""
        super().__init__(
            """
{h1}NeptuneRetryError: The Neptune server returned an error after {retries} retries, {time:.2f} seconds.{end}

{status_code_line}
{content_line}
""",
            retries=retries,
            time=time,
            status_code_line=f"Last response status: {last_status_code}" if last_status_code is not None else "",
            content_line=f"Last response content: {content_str}" if last_content is not None else "",
        )


def _decode_content(content: bytes, content_max_length: int = 1000) -> str:
    try:
        return content.decode("utf-8")[:content_max_length]
    except UnicodeDecodeError:
        return repr(content)[:content_max_length]


EMPTY_STYLES = {
    "h1": "",
    "h2": "",
    "blue": "",
    "python": "",
    "bash": "",
    "warning": "",
    "correct": "",
    "fail": "",
    "bold": "",
    "underline": "",
    "end": "",
}

UNIX_STYLES = {
    "h1": "\033[95m",
    "h2": "\033[94m",
    "blue": "\033[94m",
    "python": "\033[96m",
    "bash": "\033[95m",
    "warning": "\033[93m",
    "correct": "\033[92m",
    "fail": "\033[91m",
    "bold": "\033[1m",
    "underline": "\033[4m",
    "end": "\033[0m",
}

_styles = UNIX_STYLES if platform.system() in ["Linux", "Darwin"] else EMPTY_STYLES


def _get_styles() -> Dict[str, str]:  # TODO: unused?
    if env.NEPTUNE_ENABLE_COLORS.get():
        return _styles
    return EMPTY_STYLES


warnings.simplefilter("once", category=NeptuneWarning)

# We keep a set of types we've warned the user about to make sure we warn about a type only once.
# This is necessary because of a bug in pandas, that causes duplicate warnings to be issued everytime after an
# DataFrame() is created (presumably only empty DF).
# The bug basically makes `warnings.simplefilter("once", NeptuneWarning)` not work as expected, and would flood
# the user with warnings in some cases.
_warned_types = set()


def warn_unsupported_value_type(type_: str) -> None:
    if type_ in _warned_types:
        return

    _warned_types.add(type_)
    warnings.warn(
        f"A value of type `{type_}` was returned by your query. This type is not supported by your installed version "
        "of neptune-fetcher. Values will evaluate to `None` and empty DataFrames. "
        "Upgrade neptune-fetcher to access this data.",
        NeptuneWarning,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import pathlib
from dataclasses import dataclass
from typing import (
    Literal,
    Optional,
)

import azure.core.exceptions
from azure.storage.blob import BlobClient
from neptune_api.api.storage import signed_url
from neptune_api.client import AuthenticatedClient
from neptune_api.models import (
    CreateSignedUrlsRequest,
    CreateSignedUrlsResponse,
    FileToSign,
    Permission,
)

from .. import (
    env,
    identifiers,
)
from ..retrieval import retry


@dataclass(frozen=True)
class SignedFile:
    url: str
    path: str


def fetch_signed_urls(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    file_paths: list[str],
    permission: Literal["read", "write"] = "read",
) -> list[SignedFile]:
    body = CreateSignedUrlsRequest(
        files=[
            FileToSign(project_identifier=project_identifier, path=file_path, permission=Permission(permission))
            for file_path in file_paths
        ]
    )

    response = retry.handle_errors_default(signed_url.sync_detailed)(client=client, body=body)

    data: CreateSignedUrlsResponse = response.parsed
    return [SignedFile(url=file_.url, path=file_.path) for file_ in data.files]


def download_file(
    signed_url: str,
    target_path: pathlib.Path,
    max_concurrency: int = env.NEPTUNE_FETCHER_FILES_MAX_CONCURRENCY.get(),
    timeout: Optional[int] = env.NEPTUNE_FETCHER_FILES_TIMEOUT.get(),
) -> pathlib.Path:
    target_path.parent.mkdir(parents=True, exist_ok=True)
    with open(target_path, mode="wb") as opened:
        blob_client = BlobClient.from_blob_url(signed_url)
        download_stream = blob_client.download_blob(max_concurrency=max_concurrency, timeout=timeout)
        for chunk in download_stream.chunks():
            opened.write(chunk)
    return target_path


def download_file_retry(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    signed_file: SignedFile,
    target_path: pathlib.Path,
    retries: int = 3,
) -> Optional[pathlib.Path]:
    attempt = 0
    while True:
        try:
            return download_file(
                signed_url=signed_file.url,
                target_path=target_path,
            )
        except azure.core.exceptions.ResourceNotFoundError:
            return None
        except azure.core.exceptions.ClientAuthenticationError:
            if attempt >= retries:
                raise
            attempt += 1
            signed_file = fetch_signed_urls(
                client=client,
                project_identifier=project_identifier,
                file_paths=[signed_file.path],
            )[0]


def create_target_path(destination: pathlib.Path, experiment_name: str, attribute_path: str) -> pathlib.Path:
    relative_target_path = pathlib.Path(".") / experiment_name / attribute_path

    sanitized_parts = [_sanitize_path_part(part) for part in relative_target_path.parts]
    relative_target_path = pathlib.Path(*sanitized_parts)

    return destination / relative_target_path


def _sanitize_path_part(part: str, max_part_length: int = 255) -> str:
    # Replace invalid characters with underscores
    part = "".join(c if c.isalnum() or c in ("_", "-") else "_" for c in part)

    if len(part) > max_part_length:
        digest = hashlib.blake2b(part.encode("utf-8"), digest_size=8).hexdigest()
        part = f"{part[:max_part_length - len(digest) - 1]}_{digest}"

    return part



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import Executor
from typing import (
    Callable,
    Optional,
)

from neptune_api.client import AuthenticatedClient

from .. import (
    filters,
    identifiers,
)
from ..composition import concurrency
from ..composition.attributes import (
    AttributeDefinitionAggregation,
    fetch_attribute_definition_aggregations,
    fetch_attribute_definitions,
)
from ..retrieval import attribute_values as att_vals
from ..retrieval import (
    search,
    split,
    util,
)


def fetch_attribute_definitions_split(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    sys_ids: list[identifiers.SysId],
    downstream: Callable[[list[identifiers.SysId], util.Page[identifiers.AttributeDefinition]], concurrency.OUT],
) -> concurrency.OUT:
    return concurrency.generate_concurrently(
        items=split.split_sys_ids(sys_ids),
        executor=executor,
        downstream=lambda sys_ids_split: concurrency.generate_concurrently(
            fetch_attribute_definitions(
                client=client,
                project_identifiers=[project_identifier],
                run_identifiers=[identifiers.RunIdentifier(project_identifier, sys_id) for sys_id in sys_ids_split],
                attribute_filter=attribute_filter,
                executor=fetch_attribute_definitions_executor,
            ),
            executor=executor,
            downstream=lambda definitions: downstream(sys_ids_split, definitions),
        ),
    )


def fetch_attribute_definition_aggregations_split(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    sys_ids: list[identifiers.SysId],
    downstream: Callable[
        [
            list[identifiers.SysId],
            util.Page[identifiers.AttributeDefinition],
            util.Page[AttributeDefinitionAggregation],
        ],
        concurrency.OUT,
    ],
) -> concurrency.OUT:
    return concurrency.generate_concurrently(
        items=split.split_sys_ids(sys_ids),
        executor=executor,
        downstream=lambda sys_ids_split: concurrency.generate_concurrently(
            fetch_attribute_definition_aggregations(
                client=client,
                project_identifiers=[project_identifier],
                run_identifiers=[identifiers.RunIdentifier(project_identifier, sys_id) for sys_id in sys_ids_split],
                attribute_filter=attribute_filter,
                executor=fetch_attribute_definitions_executor,
            ),
            executor=executor,
            downstream=lambda page_pair: downstream(sys_ids_split, page_pair[0], page_pair[1]),
        ),
    )


def fetch_attribute_definitions_complete(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[filters._Filter],
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    container_type: search.ContainerType,
    downstream: Callable[[util.Page[identifiers.AttributeDefinition]], concurrency.OUT],
) -> concurrency.OUT:
    if container_type == search.ContainerType.RUN and filter_ is None:
        return concurrency.generate_concurrently(
            fetch_attribute_definitions(
                client=client,
                project_identifiers=[project_identifier],
                run_identifiers=None,
                attribute_filter=attribute_filter,
                executor=fetch_attribute_definitions_executor,
            ),
            executor=executor,
            downstream=downstream,
        )
    else:
        return concurrency.generate_concurrently(
            items=search.fetch_sys_ids(
                client=client,
                project_identifier=project_identifier,
                filter_=filter_,
                container_type=container_type,
            ),
            executor=executor,
            downstream=lambda sys_ids_page: fetch_attribute_definitions_split(
                client=client,
                project_identifier=project_identifier,
                attribute_filter=attribute_filter,
                executor=executor,
                fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                sys_ids=sys_ids_page.items,
                downstream=lambda _, definitions: downstream(definitions),
            ),
        )


def fetch_attribute_values_split(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    executor: Executor,
    sys_ids: list[identifiers.SysId],
    attribute_definitions: list[identifiers.AttributeDefinition],
    downstream: Callable[[util.Page[att_vals.AttributeValue]], concurrency.OUT],
) -> concurrency.OUT:
    return concurrency.generate_concurrently(
        items=split.split_sys_ids_attributes(sys_ids, attribute_definitions),
        executor=executor,
        downstream=lambda split_pair: concurrency.generate_concurrently(
            items=att_vals.fetch_attribute_values(
                client=client,
                project_identifier=project_identifier,
                run_identifiers=[identifiers.RunIdentifier(project_identifier, s) for s in split_pair[0]],
                attribute_definitions=split_pair[1],
            ),
            executor=executor,
            downstream=downstream,
        ),
    )



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import concurrent
from concurrent.futures import (
    Executor,
    Future,
    ThreadPoolExecutor,
)
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Optional,
    TypeVar,
)

from .. import env

T = TypeVar("T")
R = TypeVar("R")
OUT = tuple[set[Future], Optional[R]]
_Params = dict[str, Any]


def create_thread_pool_executor() -> Executor:
    max_workers = env.NEPTUNE_FETCHER_MAX_WORKERS.get()
    return ThreadPoolExecutor(max_workers=max_workers)


def generate_concurrently(
    items: Generator[T, None, None],
    executor: Executor,
    downstream: Callable[[T], OUT],
) -> OUT:
    try:
        head = next(items)
        futures = {
            executor.submit(downstream, head),
            executor.submit(generate_concurrently, items, executor, downstream),
        }
        return futures, None
    except StopIteration:
        return set(), None


def fork_concurrently(executor: Executor, downstreams: Iterable[Callable[[], OUT]]) -> OUT:
    futures = {executor.submit(downstream) for downstream in downstreams}
    return futures, None


def return_value(item: R) -> OUT:
    return set(), item


def gather_results(output: OUT) -> Generator[R, None, None]:
    futures, value = output
    if value is not None:
        yield value
    while futures:
        done, not_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
        futures = not_done
        for future in done:
            new_futures, value = future.result()
            futures.update(new_futures)
            if value is not None:
                yield value



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import Executor
from dataclasses import dataclass
from typing import (
    Generator,
    Iterable,
    Literal,
    Optional,
)

from neptune_api.client import AuthenticatedClient

from .. import (
    env,
    filters,
    identifiers,
)
from ..composition import concurrency
from ..retrieval import attribute_definitions as att_defs
from ..retrieval import util
from ..retrieval.attribute_types import TYPE_AGGREGATIONS


@dataclass(frozen=True)
class AttributeDefinitionAggregation:
    attribute_definition: identifiers.AttributeDefinition
    aggregation: Literal["last", "min", "max", "average", "variance"]


def fetch_attribute_definitions(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    batch_size: int = env.NEPTUNE_FETCHER_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
    pages_filters = _fetch_attribute_definitions(
        client, project_identifiers, run_identifiers, attribute_filter, batch_size, executor
    )

    seen_items: set[identifiers.AttributeDefinition] = set()
    for page, filter_ in pages_filters:
        new_items = [item for item in page.items if item not in seen_items]
        seen_items.update(new_items)
        yield util.Page(items=new_items)


def fetch_attribute_definition_aggregations(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Iterable[identifiers.RunIdentifier],
    attribute_filter: filters._BaseAttributeFilter,
    executor: Executor,
    batch_size: int = env.NEPTUNE_FETCHER_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[
    tuple[util.Page[identifiers.AttributeDefinition], util.Page[AttributeDefinitionAggregation]], None, None
]:
    """
    Each attribute definition is yielded once when it's first encountered.
    If the attribute definition is of a type that supports aggregations (for now only float_series),
    it's then yielded once for each aggregation in the filter that returned it.
    """

    pages_filters = _fetch_attribute_definitions(
        client, project_identifiers, run_identifiers, attribute_filter, batch_size, executor
    )

    seen_definitions: set[identifiers.AttributeDefinition] = set()
    seen_definition_aggregations: set[AttributeDefinitionAggregation] = set()

    for page, filter_ in pages_filters:
        new_definitions = []
        new_definition_aggregations = []

        for definition in page.items:
            if definition not in seen_definitions:
                new_definitions.append(definition)
                seen_definitions.add(definition)

            if definition.type in TYPE_AGGREGATIONS.keys():
                for aggregation in filter_.aggregations:
                    if aggregation not in TYPE_AGGREGATIONS[definition.type]:
                        continue

                    definition_aggregation = AttributeDefinitionAggregation(
                        attribute_definition=definition, aggregation=aggregation
                    )
                    if definition_aggregation not in seen_definition_aggregations:
                        new_definition_aggregations.append(definition_aggregation)
                        seen_definition_aggregations.add(definition_aggregation)

        yield util.Page(items=new_definitions), util.Page(items=new_definition_aggregations)


def _fetch_attribute_definitions(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._BaseAttributeFilter,
    batch_size: int,
    executor: Executor,
) -> Generator[tuple[util.Page[identifiers.AttributeDefinition], filters._AttributeFilter], None, None]:
    def go_fetch_single(
        filter_: filters._AttributeFilter,
    ) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
        return att_defs.fetch_attribute_definitions_single_filter(
            client=client,
            project_identifiers=project_identifiers,
            run_identifiers=run_identifiers,
            attribute_filter=filter_,
            batch_size=batch_size,
        )

    filters_ = att_defs.split_attribute_filters(attribute_filter)

    if len(filters_) == 1:
        head = filters_[0]
        for page in go_fetch_single(head):
            yield page, head
    else:
        output = concurrency.generate_concurrently(
            items=(filter_ for filter_ in filters_),
            executor=executor,
            downstream=lambda filter_: concurrency.generate_concurrently(
                items=go_fetch_single(filter_),
                executor=executor,
                downstream=lambda _page: concurrency.return_value((_page, filter_)),
            ),
        )
        yield from concurrency.gather_results(output)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
)

from neptune_api.proto.neptune_pb.api.v1.model.leaderboard_entries_pb2 import (
    ProtoAttributeDTO,
    ProtoFileRefAttributeDTO,
    ProtoFileRefSeriesAttributeDTO,
    ProtoFloatSeriesAttributeDTO,
    ProtoHistogramSeriesAttributeDTO,
    ProtoStringSeriesAttributeDTO,
)

from ...exceptions import warn_unsupported_value_type

ALL_TYPES = (
    "float",
    "int",
    "string",
    "bool",
    "datetime",
    "float_series",
    "string_set",
    "string_series",
    "file",
    "file_series",
    "histogram_series",
)
FLOAT_SERIES_AGGREGATIONS = frozenset({"last", "min", "max", "average", "variance"})
STRING_SERIES_AGGREGATIONS = frozenset({"last"})
FILE_SERIES_AGGREGATIONS = frozenset({"last"})
HISTOGRAM_SERIES_AGGREGATIONS = frozenset({"last"})
ALL_AGGREGATIONS = (
    FLOAT_SERIES_AGGREGATIONS | STRING_SERIES_AGGREGATIONS | FILE_SERIES_AGGREGATIONS | HISTOGRAM_SERIES_AGGREGATIONS
)
TYPE_AGGREGATIONS = {
    "float_series": FLOAT_SERIES_AGGREGATIONS,
    "string_series": STRING_SERIES_AGGREGATIONS,
    "file_series": FILE_SERIES_AGGREGATIONS,
    "histogram_series": HISTOGRAM_SERIES_AGGREGATIONS,
}

_ATTRIBUTE_TYPE_PYTHON_TO_BACKEND_MAP = {
    "float_series": "floatSeries",
    "string_set": "stringSet",
    "string_series": "stringSeries",
    "file": "fileRef",
    "file_series": "fileRefSeries",
    "histogram_series": "histogramSeries",
}

_ATTRIBUTE_TYPE_BACKEND_TO_PYTHON_MAP = {v: k for k, v in _ATTRIBUTE_TYPE_PYTHON_TO_BACKEND_MAP.items()}


def map_attribute_type_python_to_backend(_type: str) -> str:
    return _ATTRIBUTE_TYPE_PYTHON_TO_BACKEND_MAP.get(_type, _type)


def map_attribute_type_backend_to_python(_type: str) -> str:
    return _ATTRIBUTE_TYPE_BACKEND_TO_PYTHON_MAP.get(_type, _type)


@dataclass(frozen=True)
class FloatSeriesAggregations:
    last: float
    min: float
    max: float
    average: float
    variance: float


@dataclass(frozen=True)
class File:
    path: str
    size_bytes: int
    mime_type: str

    def __repr__(self):
        return f"File({self.mime_type}, size={humanize_size(self.size_bytes)})"


@dataclass(frozen=True)
class Histogram:
    type: str
    edges: list[float]
    values: list[float]


@dataclass(frozen=True)
class StringSeriesAggregations:
    last: Optional[str]
    last_step: Optional[float]


@dataclass(frozen=True)
class FileSeriesAggregations:
    last: Optional[File]
    last_step: Optional[float]


@dataclass(frozen=True)
class HistogramSeriesAggregations:
    last: Optional[Histogram]
    last_step: Optional[float]


def humanize_size(size_bytes: int) -> str:
    """Convert bytes to a human-readable format."""
    if size_bytes < 1024:
        return f"{size_bytes} B"
    elif size_bytes < 1024**2:
        return f"{size_bytes / 1024:.2f} KB"
    elif size_bytes < 1024**3:
        return f"{size_bytes / (1024 ** 2):.2f} MB"
    else:
        return f"{size_bytes / (1024 ** 3):.2f} GB"


def extract_value(attr: ProtoAttributeDTO) -> Optional[Any]:
    if attr.type == "floatSeries":
        return _extract_float_series_aggregations(attr.float_series_properties)
    elif attr.type == "stringSeries":
        return _extract_string_series_aggregations(attr.string_series_properties)
    elif attr.type == "fileRefSeries":
        return _extract_file_ref_series_aggregations(attr.file_ref_series_properties)
    elif attr.type == "histogramSeries":
        return _extract_histogram_series_aggregations(attr.histogram_series_properties)
    elif attr.type == "string":
        return attr.string_properties.value
    elif attr.type == "int":
        return attr.int_properties.value
    elif attr.type == "float":
        return attr.float_properties.value
    elif attr.type == "bool":
        return attr.bool_properties.value
    elif attr.type == "datetime":
        return datetime.datetime.fromtimestamp(attr.datetime_properties.value / 1000, tz=datetime.timezone.utc)
    elif attr.type == "stringSet":
        return set(attr.string_set_properties.value)
    elif attr.type == "fileRef":
        return _extract_file_ref_properties(attr.file_ref_properties)
    elif attr.type == "experimentState":
        return None
    else:
        warn_unsupported_value_type(attr.type)
        return None


def _extract_float_series_aggregations(attr: ProtoFloatSeriesAttributeDTO) -> FloatSeriesAggregations:
    return FloatSeriesAggregations(
        last=attr.last,
        min=attr.min,
        max=attr.max,
        average=attr.average,
        variance=attr.variance,
    )


def _extract_file_ref_properties(attr: ProtoFileRefAttributeDTO) -> File:
    return File(
        path=attr.path,
        size_bytes=attr.sizeBytes,
        mime_type=attr.mimeType,
    )


def _extract_string_series_aggregations(attr: ProtoStringSeriesAttributeDTO) -> StringSeriesAggregations:
    return StringSeriesAggregations(
        last=attr.last,
        last_step=attr.last_step,
    )


def _extract_file_ref_series_aggregations(attr: ProtoFileRefSeriesAttributeDTO) -> FileSeriesAggregations:
    return FileSeriesAggregations(
        last=File(
            path=attr.last.path,
            size_bytes=attr.last.sizeBytes,
            mime_type=attr.last.mimeType,
        ),
        last_step=attr.last_step,
    )


def _extract_histogram_series_aggregations(attr: ProtoHistogramSeriesAttributeDTO) -> HistogramSeriesAggregations:
    return HistogramSeriesAggregations(
        last=Histogram(
            type=attr.last.type,
            edges=list(attr.last.edges),
            values=list(attr.last.values),
        ),
        last_step=attr.last_step,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    "list_runs",
    "list_attributes",
    "fetch_runs_table",
    "fetch_metrics",
    "fetch_series",
]

from typing import (
    Literal,
    Optional,
    Tuple,
    Union,
)

import pandas as _pandas

from neptune_query import filters
from neptune_query._internal import (
    get_default_project_identifier,
    resolve_attributes_filter,
    resolve_runs_filter,
    resolve_sort_by,
)
from neptune_query.internal.composition import fetch_metrics as _fetch_metrics
from neptune_query.internal.composition import fetch_series as _fetch_series
from neptune_query.internal.composition import fetch_table as _fetch_table
from neptune_query.internal.composition import list_attributes as _list_attributes
from neptune_query.internal.composition import list_containers as _list_containers
from neptune_query.internal.retrieval import search as _search


def list_runs(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
) -> list[str]:
    """
     Returns a list of run IDs in a project.

    `project` - the project name to use; if not provided, NEPTUNE_PROJECT env var is used
    `runs` - a filter specifying which runs to include
        - a list of specific run IDs, or
        - a regex that the run ID must match, or
        - a Filter object
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)

    return _list_containers.list_containers(
        project_identifier=project_identifier,
        filter_=runs_filter,
        container_type=_search.ContainerType.RUN,
    )


def list_attributes(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Optional[Union[str, list[str], filters.AttributeFilter]] = None,
) -> list[str]:
    """
    List the names of attributes in a project.
    Optionally filter by runs and attributes.

    `project` - the project name to use; if not provided, NEPTUNE_PROJECT env var is used
    `runs` - a filter specifying runs to which the attributes belong
        - a list of specific run IDs, or
        - a regex that the run ID must match, or
        - a Filter object
    `attributes` - a filter specifying which attributes to include in the table
        - a list of specific attribute names, or
        - a regex that the attribute name must match, or
        - an AttributeFilter object;

    Returns a list of unique attribute names in runs matching the filter.
    """

    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _list_attributes.list_attributes(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        container_type=_search.ContainerType.RUN,
    )


def fetch_metrics(
    *,
    project: Optional[str] = None,
    runs: Union[str, list[str], filters.Filter],
    attributes: Union[str, list[str], filters.AttributeFilter],
    include_time: Optional[Literal["absolute"]] = None,
    step_range: Tuple[Optional[float], Optional[float]] = (None, None),
    lineage_to_the_root: bool = True,
    tail_limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
    include_point_previews: bool = False,
) -> _pandas.DataFrame:
    """
    Returns raw values for the requested metrics (no aggregation, approximation, or interpolation).

    `project` - the project name to use; if not provided, NEPTUNE_PROJECT env var is used
    `runs` - a filter specifying which runs to include
        - a list of specific run IDs, or
        - a regex that the run ID must match, or
        - a Filter object
    `attributes` - a filter specifying which attributes to include in the table
        - a list of specific attribute names, or
        - a regex that the attribute name must match, or
        - an AttributeFilter object;
    `include_time` - whether to include absolute timestamp
    `step_range` - a tuple specifying the range of steps to include; can represent an open interval
    `lineage_to_the_root` - if True (default), includes all points from the complete run history.
        If False, only includes points from the most recent run in the lineage.
    `tail_limit` - from the tail end of each series, how many points to include at most.
    `type_suffix_in_column_names` - False by default. If set to True, columns of the returned DataFrame
        are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
        If False, an exception is raised if there are multiple types under one attribute path.
    `include_point_previews` - False by default. If False the returned results will only contain committed
        points. If True the results will also include preview points and the returned DataFrame will
        have additional sub-columns with preview status (is_preview and preview_completion).

    If `include_time` is set, each metric column has an additional sub-column with requested timestamp values.
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _fetch_metrics.fetch_metrics(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        include_time=include_time,
        step_range=step_range,
        lineage_to_the_root=lineage_to_the_root,
        tail_limit=tail_limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        include_point_previews=include_point_previews,
        container_type=_search.ContainerType.RUN,
    )


def fetch_runs_table(
    *,
    project: Optional[str] = None,
    runs: Optional[Union[str, list[str], filters.Filter]] = None,
    attributes: Union[str, list[str], filters.AttributeFilter] = "^sys/name$",
    sort_by: Union[str, filters.Attribute] = filters.Attribute("sys/creation_time", type="datetime"),
    sort_direction: Literal["asc", "desc"] = "desc",
    limit: Optional[int] = None,
    type_suffix_in_column_names: bool = False,
) -> _pandas.DataFrame:
    """
    `project` - the project name to use; if not provided, NEPTUNE_PROJECT env var is used
    `runs` - a filter specifying which runs to include in the table
        - a list of specific run IDs, or
        - a regex that the run ID must match, or
        - a Filter object
    `attributes` - a filter specifying which attributes to include in the table
        - a list of specific attribute names, or
        - a regex that the attribute name must match, or
        - an AttributeFilter object
    `sort_by` - an attribute name or an Attribute object specifying type
    `sort_direction` - 'asc' or 'desc'
    `limit` - maximum number of runs to return; by default all runs are returned.
    `type_suffix_in_column_names` - False by default. If set to True, columns of the returned DataFrame
        are suffixed with ":<type>", e.g. "attribute1:float_series", "attribute1:string".
        If False, an exception is raised if there are multiple types under one attribute path.

    Returns a DataFrame similar to the Runs Table in the UI.
    (Only the last logged value of each metric is returned, no aggregations or approximations)
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)
    sort_by = resolve_sort_by(sort_by)

    return _fetch_table.fetch_table(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        sort_by=sort_by,
        sort_direction=sort_direction,
        limit=limit,
        type_suffix_in_column_names=type_suffix_in_column_names,
        container_type=_search.ContainerType.RUN,
        flatten_aggregations=True,
    )


def fetch_series(
    *,
    project: Optional[str] = None,
    runs: Union[str, list[str], filters.Filter],
    attributes: Union[str, list[str], filters.AttributeFilter],
    include_time: Optional[Literal["absolute"]] = None,
    step_range: Tuple[Optional[float], Optional[float]] = (None, None),
    lineage_to_the_root: bool = True,
    tail_limit: Optional[int] = None,
) -> _pandas.DataFrame:
    """
    Fetches raw values for string series from selected runs.

    Currently only supports attributes of type string_series.

    `project` - the project name to use; if not provided, NEPTUNE_PROJECT env var is used
    `runs` - a filter specifying which runs to include
        - a list of specific run IDs, or
        - a regex that experiment name must match, or
        - a Filter object for more complex filtering
    `attributes` - a filter specifying which attributes to include
        - a list of specific attribute names, or
        - a regex that the attribute name must match, or
        - an AttributeFilter object;
    `include_time` - whether to include absolute timestamp
    `step_range` - tuple specifying the range of steps to include; can represent an open interval
    `lineage_to_the_root` - if True (default), includes all points from the complete experiment history.
        If False, only includes points from the most recent experiment in the lineage.
    `tail_limit` - from the tail end of each series, maximum number of points to include.

    Returns a DataFrame containing string series for the specified runs and attributes.
    If include_time is set, each series column will have an additional sub-column with the requested timestamp values.
    """
    project_identifier = get_default_project_identifier(project)
    runs_filter = resolve_runs_filter(runs)
    attributes_filter = resolve_attributes_filter(attributes)

    return _fetch_series.fetch_series(
        project_identifier=project_identifier,
        filter_=runs_filter,
        attributes=attributes_filter,
        include_time=include_time,
        step_range=step_range,
        lineage_to_the_root=lineage_to_the_root,
        tail_limit=tail_limit,
        container_type=_search.ContainerType.RUN,
    )



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

__all__ = ("get_client", "clear_cache")

import logging
import threading
from typing import (
    Dict,
    Optional,
    Tuple,
)

from neptune_api import AuthenticatedClient
from neptune_api.credentials import Credentials

from .api_utils import (
    create_auth_api_client,
    get_config_and_token_urls,
)
from .context import Context

# Disable httpx logging, httpx logs requests at INFO level
logging.getLogger("httpx").setLevel(logging.WARN)

_cache: dict[int, AuthenticatedClient] = {}
_lock: threading.RLock = threading.RLock()


def get_client(context: Context, proxies: Optional[Dict[str, str]] = None) -> AuthenticatedClient:
    hash_key = hash((context.api_token, _dict_to_hashable(proxies)))

    # We hold the lock during the entire client creation process, including networking.
    # This is a deliberate decision to make sure that only one client is created for a given hash_key,
    # at any given time.
    # The process can be optimized later by more granular locking (per hash_key), but for the time being
    # the decision is to keep it dead simple.

    with _lock:
        if (client := _cache.get(hash_key)) is not None:
            return client

        credentials = Credentials.from_api_key(api_key=context.api_token)
        config, token_urls = get_config_and_token_urls(credentials=credentials, proxies=proxies)
        client = create_auth_api_client(
            credentials=credentials,
            config=config,
            token_refreshing_urls=token_urls,
            proxies=proxies,
        )

        _cache[hash_key] = client
        return client


def clear_cache() -> None:
    with _lock:
        _cache.clear()


def _dict_to_hashable(d: Optional[Dict[str, str]]) -> frozenset[Tuple[str, str]]:
    """Convert a dict to a hashable data structure"""

    items = tuple() if d is None else d.items()
    return frozenset(items)



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Collection,
    Optional,
    Sequence,
    Union,
)


def _validate_string_or_string_list(value: Optional[Union[str, list[str]]], field_name: str) -> None:
    """Validate that a value is either None, a string, or a list of strings."""
    if value is not None:
        if isinstance(value, list):
            if not all(isinstance(item, str) for item in value):
                raise ValueError(f"{field_name} must be a string or list of strings")
        elif not isinstance(value, str):
            raise ValueError(f"{field_name} must be a string or list of strings")


def _validate_string_list(value: Optional[list[str]], field_name: str, disallow_empty: bool = False) -> None:
    """Validate that a value is either None or a list of strings."""
    msg = f"{field_name} must not be a non-empty list" if disallow_empty else f"{field_name} must be a list of strings"
    if value is not None:
        if not isinstance(value, list) or not all(isinstance(item, str) for item in value):
            raise ValueError(msg)
        if disallow_empty and len(value) == 0:
            raise ValueError(msg)


def _validate_list_of_allowed_values(value: Sequence[str], allowed_values: Collection[str], field_name: str) -> None:
    """Validate that a value is a list containing only allowed values."""
    if not isinstance(value, Sequence) or not all(isinstance(v, str) and v in allowed_values for v in value):
        raise ValueError(f"{field_name} must be a list of valid values: {sorted(allowed_values)} (got {value!r})")


def _validate_allowed_value(value: Optional[str], allowed_values: Collection[str], field_name: str) -> None:
    """Validate that a value is either None or one of the allowed values."""
    if value is not None and (not isinstance(value, str) or value not in allowed_values):
        raise ValueError(f"{field_name} must be one of: {sorted(allowed_values)}")



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools as ft
import itertools as it
import re
from typing import (
    Any,
    Generator,
    Iterable,
    Optional,
    Union,
)

from neptune_api.api.retrieval import query_attribute_definitions_within_project
from neptune_api.client import AuthenticatedClient
from neptune_api.models import (
    QueryAttributeDefinitionsBodyDTO,
    QueryAttributeDefinitionsResultDTO,
)

from .. import filters  # noqa: E402
from .. import (  # noqa: E402
    env,
    identifiers,
)
from ..retrieval import attribute_types as types  # noqa: E402
from ..retrieval import util  # noqa: E402
from ..retrieval import retry


def split_attribute_filters(
    _attribute_filter: filters._BaseAttributeFilter,
) -> list[filters._AttributeFilter]:
    if isinstance(_attribute_filter, filters._AttributeFilter):
        return [_attribute_filter]
    elif isinstance(_attribute_filter, filters._AttributeFilterAlternative):
        return list(it.chain.from_iterable(split_attribute_filters(child) for child in _attribute_filter.filters))
    else:
        raise RuntimeError(f"Unexpected filter type: {type(_attribute_filter)}")


def fetch_attribute_definitions_single_filter(
    client: AuthenticatedClient,
    project_identifiers: Iterable[identifiers.ProjectIdentifier],
    run_identifiers: Optional[Iterable[identifiers.RunIdentifier]],
    attribute_filter: filters._AttributeFilter,
    batch_size: int = env.NEPTUNE_FETCHER_ATTRIBUTE_DEFINITIONS_BATCH_SIZE.get(),
) -> Generator[util.Page[identifiers.AttributeDefinition], None, None]:
    params: dict[str, Any] = {
        "projectIdentifiers": list(project_identifiers),
        "attributeNameFilter": dict(),
        "nextPage": {"limit": batch_size},
    }

    if run_identifiers is not None:
        params["experimentIdsFilter"] = [str(e) for e in run_identifiers]

    # Convert name_eq to an additional condition added to each of must_match_any alternatives.
    name_regexes = None
    if attribute_filter.name_eq is not None:
        name_regexes = _escape_name_eq(_variants_to_list(attribute_filter.name_eq))

    if attribute_filter.must_match_any is not None:
        attribute_name_filter_dtos = []
        for alternative in attribute_filter.must_match_any:
            attribute_name_filter_dto = {}
            must_match_regexes = _union_options([name_regexes, alternative.must_match_regexes])
            if must_match_regexes is not None:
                attribute_name_filter_dto["mustMatchRegexes"] = must_match_regexes
            if alternative.must_not_match_regexes is not None:
                attribute_name_filter_dto["mustNotMatchRegexes"] = alternative.must_not_match_regexes
            if attribute_name_filter_dto:
                attribute_name_filter_dtos.append(attribute_name_filter_dto)
        params["attributeNameFilter"]["mustMatchAny"] = attribute_name_filter_dtos

    elif name_regexes is not None:
        params["attributeNameFilter"]["mustMatchAny"] = [{"mustMatchRegexes": name_regexes}]

    attribute_types = _variants_to_list(attribute_filter.type_in)
    if attribute_types is not None:
        params["attributeFilter"] = [
            {"attributeType": types.map_attribute_type_python_to_backend(_type)} for _type in attribute_types
        ]

    # note: attribute_filter.aggregations is intentionally ignored

    return util.fetch_pages(
        client=client,
        fetch_page=_fetch_attribute_definitions_page,
        process_page=_process_attribute_definitions_page,
        make_new_page_params=ft.partial(_make_new_attribute_definitions_page_params, batch_size=batch_size),
        params=params,
    )


def _fetch_attribute_definitions_page(
    client: AuthenticatedClient,
    params: dict[str, Any],
) -> QueryAttributeDefinitionsResultDTO:
    body = QueryAttributeDefinitionsBodyDTO.from_dict(params)

    response = retry.handle_errors_default(query_attribute_definitions_within_project.sync_detailed)(
        client=client,
        body=body,
    )

    return response.parsed


def _process_attribute_definitions_page(
    data: QueryAttributeDefinitionsResultDTO,
) -> util.Page[identifiers.AttributeDefinition]:
    items = []
    for entry in data.entries:
        item = identifiers.AttributeDefinition(
            name=entry.name,
            type=types.map_attribute_type_backend_to_python(str(entry.type)),
        )
        items.append(item)
    return util.Page(items=items)


def _make_new_attribute_definitions_page_params(
    params: dict[str, Any],
    data: Optional[QueryAttributeDefinitionsResultDTO],
    batch_size: int,
) -> Optional[dict[str, Any]]:
    if data is None:
        if "nextPageToken" in params["nextPage"]:
            del params["nextPage"]["nextPageToken"]
        return params

    next_page_token = data.next_page.next_page_token
    if not next_page_token or len(data.entries) < batch_size:
        return None

    params["nextPage"]["nextPageToken"] = next_page_token
    return params


def _escape_name_eq(names: Optional[list[str]]) -> Optional[list[str]]:
    if names is None:
        return None

    escaped = [f"{re.escape(name)}" for name in names]

    if len(escaped) == 1:
        return [f"^{escaped[0]}$"]
    else:
        joined = "|".join(escaped)
        return [f"^({joined})$"]


def _variants_to_list(param: Union[str, Iterable[str], None]) -> Optional[list[str]]:
    if param is None:
        return None
    if isinstance(param, str):
        return [param]
    return list(param)


def _union_options(options: list[Optional[list[str]]]) -> Optional[list[str]]:
    result = None

    for option in options:
        if option is not None:
            if result is None:
                result = []
            result.extend(option)

    return result



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from abc import ABC
from dataclasses import (
    dataclass,
    field,
)
from datetime import datetime
from typing import (
    Callable,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Union,
)

from .retrieval import attribute_types as types

__all__ = ["_Filter", "_AttributeFilter", "_Attribute"]

from .util import (
    _validate_allowed_value,
    _validate_list_of_allowed_values,
    _validate_string_list,
    _validate_string_or_string_list,
)

ATTRIBUTE_LITERAL = Literal[
    "float",
    "int",
    "string",
    "bool",
    "datetime",
    "float_series",
    "string_set",
    "string_series",
    "file",
    "file_series",
    "histogram_series",
]
AGGREGATION_LITERAL = Literal["last", "min", "max", "average", "variance"]


class _BaseAttributeFilter(ABC):
    @staticmethod
    def any(filters: list["_BaseAttributeFilter"]) -> "_BaseAttributeFilter":
        return _AttributeFilterAlternative(filters=filters)

    @abc.abstractmethod
    def transform(
        self, map_attribute_filter: Callable[["_AttributeFilter"], "_AttributeFilter"]
    ) -> "_BaseAttributeFilter":
        ...


@dataclass
class _AttributeNameFilter:
    must_match_regexes: Optional[list[str]] = None
    must_not_match_regexes: Optional[list[str]] = None


@dataclass
class _AttributeFilter(_BaseAttributeFilter):
    """Filter to apply to attributes when fetching runs or experiments.

    Use to select specific metrics or other metadata based on various criteria.

    Args:
        name_eq (Union[str, list[str], None]): An attribute name or list of names to match exactly.
            If `None`, this filter is not applied.
        type_in (list[Literal["float", "int", "string", "bool", "datetime", "float_series", "string_set",
        "string_series", "file"]]):
            A list of allowed attribute types. Defaults to all available types.
            For a reference, see: https://docs.neptune.ai/attribute_types
        must_match_any (Optional[list[_AttributeNameFilter]]):
            If `None`, this filter is not applied. Otherwise, it's a list of `_AttributeNameFilter` instances.
            Each instance specifies a set of regexes that the attribute name must match or not match.
            Any one of the list items must match for the attribute to be included.
            Each list item should be an `_AttributeNameFilter` and it can contain:
                name_matches_all (Union[str, list[str], None]): A list of regular expressions that the attribute
                    name must match. If `None`, this filter is not applied.
                name_matches_none (Union[str, list[str], None]): A list of regular expressions that the attribute
                    names mustn't match. Attributes matching any of the regexes are excluded.
                    If `None`, this filter is not applied.
        aggregations (list[Literal["last", "min", "max", "average", "variance"]]): List of
            aggregation functions to apply when fetching metrics of type FloatSeries or StringSeries.
            Defaults to ["last"].

    Example:

    ```
    from .filters import _AttributeFilter


    loss_avg_and_var = _AttributeFilter(
        type_in=["float_series"],
        name_matches_all=[r"loss$"],
        aggregations=["average", "variance"],
    )

    npt.fetch_experiments_table(attributes=loss_avg_and_var)
    ```
    """

    name_eq: Union[str, list[str], None] = None
    type_in: Sequence[ATTRIBUTE_LITERAL] = field(default_factory=lambda: list(types.ALL_TYPES))
    must_match_any: Optional[list[_AttributeNameFilter]] = None
    aggregations: Sequence[AGGREGATION_LITERAL] = field(default_factory=lambda: ["last"])

    def __post_init__(self) -> None:
        _validate_string_or_string_list(self.name_eq, "name_eq")

        if self.must_match_any is not None:
            if not isinstance(self.must_match_any, list):
                raise ValueError("must_match_any must be a list of _AttributeNameFilter instances")
            for item in self.must_match_any:
                if not isinstance(item, _AttributeNameFilter):
                    raise ValueError("must_match_any must contain only _AttributeNameFilter instances")
                _validate_string_list(item.must_match_regexes, "must_match_regexes")
                _validate_string_list(item.must_not_match_regexes, "must_not_match_regexes")

        _validate_list_of_allowed_values(self.type_in, types.ALL_TYPES, "type_in")
        _validate_list_of_allowed_values(self.aggregations, types.ALL_AGGREGATIONS, "aggregations")

    def transform(
        self, map_attribute_filter: Callable[["_AttributeFilter"], "_AttributeFilter"]
    ) -> "_BaseAttributeFilter":
        return map_attribute_filter(self)


@dataclass
class _AttributeFilterAlternative(_BaseAttributeFilter):
    filters: Iterable[_BaseAttributeFilter]

    def __or__(self, other: "_BaseAttributeFilter") -> "_BaseAttributeFilter":
        filters = tuple(self.filters) + (other,)
        return _AttributeFilterAlternative(filters)

    def transform(
        self, map_attribute_filter: Callable[["_AttributeFilter"], "_AttributeFilter"]
    ) -> "_BaseAttributeFilter":
        transformed_filters = [f.transform(map_attribute_filter) for f in self.filters]
        return _AttributeFilterAlternative(transformed_filters)


@dataclass
class _Attribute:
    """Helper for specifying an attribute and picking a metric aggregation function.

    When fetching experiments or runs, use this class to filter and sort the returned entries.

    Args:
        name (str): An attribute name to match exactly.
        aggregation (Literal["last", "min", "max", "average", "variance"], optional):
            Aggregation function to apply when specifying a metric of type FloatSeries.
            Defaults to `"last"`, i.e. the last logged value.
        type (Literal["float", "int", "string", "bool", "datetime", "float_series", "string_set"], optional):
            Attribute type. Specify it to resolve ambiguity, in case some of the project's runs contain attributes
            that have the same name but are of a different type.
            For a reference, see: https://docs.neptune.ai/attribute_types

    Example:

    Select a metric and pick variance as the aggregation:

    ```
    from .filters import _Attribute, _Filter


    val_loss_variance = _Attribute(
        name="val/loss",
        aggregation="variance",
    )
    # Construct a filter and pass it to a fetching or listing method
    tiny_val_loss_variance = _Filter.lt(val_loss_variance, 0.01)
    npt.fetch_experiments_table(experiments=tiny_val_loss_variance)
    ```
    """

    name: str
    aggregation: Optional[AGGREGATION_LITERAL] = None
    type: Optional[ATTRIBUTE_LITERAL] = None

    def __post_init__(self) -> None:
        _validate_allowed_value(self.aggregation, types.ALL_AGGREGATIONS, "aggregation")
        _validate_allowed_value(self.type, types.ALL_TYPES, "type")  # type: ignore

    def to_query(self) -> str:
        query = f"`{self.name}`"

        if self.type is not None:
            _type = types.map_attribute_type_python_to_backend(self.type)
            query += f":{_type}"

        if self.aggregation is not None:
            query = f"{self.aggregation}({query})"

        return query

    def __str__(self) -> str:
        return self.to_query()


class _Filter(ABC):
    """Filter used to specify criteria when fetching experiments or runs.

    Examples of filters:
        - Name or attribute must match regular expression.
        - Attribute value must pass a condition, like "greater than 0.9".

    You can negate a filter or join multiple filters with logical operators.

    Methods available for attribute values:
    - `name_eq()`: Run or experiment name equals
    - `name_in()`: Run or experiment name equals any of the provided names
    - `eq()`: Value equals
    - `ne()`: Value doesn't equal
    - `gt()`: Value is greater than
    - `ge()`: Value is greater than or equal to
    - `lt()`: Value is less than
    - `le()`: Value is less than or equal to
    - `matches_all()`: Value matches regex or all in list of regexes
    - `matches_none()`: Value doesn't match regex or any of list of regexes
    - `contains_all()`: Tagset contains all tags, or string contains substrings
    - `contains_none()`: Tagset doesn't contain any of the tags, or string doesn't contain the substrings
    - `exists()`: Attribute exists

    Examples:

    ```
    from .filters import _Filter

    # Fetch metadata from specific experiments
    specific_experiments = _Filter.name_in("flying-123", "swimming-77")
    npt.fetch_experiments_table(experiments=specific_experiments)

    # Define various criteria
    owned_by_me = _Filter.eq("sys/owner", "vidar")
    loss_filter = _Filter.lt("validation/loss", 0.1)
    tag_filter = _Filter.contains_none("sys/tags", ["test", "buggy"])
    dataset_check = _Filter.exists("dataset_version")

    my_interesting_experiments = owned_by_me & loss_filter & tag_filter & dataset_check
    npt.fetch_experiments_table(experiments=my_interesting_experiments)
    ```
    """

    @staticmethod
    def eq(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator="==", attribute=attribute, value=value)

    @staticmethod
    def ne(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator="!=", attribute=attribute, value=value)

    @staticmethod
    def gt(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator=">", attribute=attribute, value=value)

    @staticmethod
    def ge(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator=">=", attribute=attribute, value=value)

    @staticmethod
    def lt(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator="<", attribute=attribute, value=value)

    @staticmethod
    def le(attribute: Union[str, _Attribute], value: Union[int, float, str, datetime]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributeValuePredicate(operator="<=", attribute=attribute, value=value)

    @staticmethod
    def matches_all(attribute: Union[str, _Attribute], regex: Union[str, list[str]]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        if isinstance(regex, str):
            return _AttributeValuePredicate(operator="MATCHES", attribute=attribute, value=regex)
        else:
            filters = [_Filter.matches_all(attribute, r) for r in regex]
            return _Filter.all(filters)

    @staticmethod
    def matches_none(attribute: Union[str, _Attribute], regex: Union[str, list[str]]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        if isinstance(regex, str):
            return _AttributeValuePredicate(operator="NOT MATCHES", attribute=attribute, value=regex)
        else:
            filters = [_Filter.matches_none(attribute, r) for r in regex]
            return _Filter.all(filters)

    @staticmethod
    def contains_all(attribute: Union[str, _Attribute], value: Union[str, list[str]]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        if isinstance(value, str):
            return _AttributeValuePredicate(operator="CONTAINS", attribute=attribute, value=value)
        else:
            filters = [_Filter.contains_all(attribute, v) for v in value]
            return _Filter.all(filters)

    @staticmethod
    def contains_none(attribute: Union[str, _Attribute], value: Union[str, list[str]]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        if isinstance(value, str):
            return _AttributeValuePredicate(operator="NOT CONTAINS", attribute=attribute, value=value)
        else:
            filters = [_Filter.contains_none(attribute, v) for v in value]
            return _Filter.all(filters)

    @staticmethod
    def exists(attribute: Union[str, _Attribute]) -> "_Filter":
        if isinstance(attribute, str):
            attribute = _Attribute(name=attribute)
        return _AttributePredicate(postfix_operator="EXISTS", attribute=attribute)

    @staticmethod
    def all(filters: list["_Filter"]) -> "_Filter":
        """Combine multiple filters with a logical AND operator."""
        if not filters or not isinstance(filters, list):
            raise ValueError("At least one filter must be provided to combine with AND.")
        return _AssociativeOperator(operator="AND", filters=filters)

    @staticmethod
    def any(filters: list["_Filter"]) -> "_Filter":
        """Combine multiple filters with a logical OR operator."""
        if not filters or not isinstance(filters, list):
            raise ValueError("At least one filter must be provided to combine with OR.")
        return _AssociativeOperator(operator="OR", filters=filters)

    @staticmethod
    def negate(filter_: "_Filter") -> "_Filter":
        return _PrefixOperator(operator="NOT", filter_=filter_)

    @staticmethod
    def name_eq(name: str) -> "_Filter":
        name_attribute = _Attribute(name="sys/name", type="string")
        return _Filter.eq(name_attribute, name)

    @abc.abstractmethod
    def to_query(self) -> str:
        ...

    def __str__(self) -> str:
        return self.to_query()


@dataclass
class _AttributeValuePredicate(_Filter):
    operator: Literal["==", "!=", ">", ">=", "<", "<=", "MATCHES", "NOT MATCHES", "CONTAINS", "NOT CONTAINS"]
    attribute: _Attribute
    value: Union[bool, int, float, str, datetime]

    def __post_init__(self) -> None:
        allowed_operators = {"==", "!=", ">", ">=", "<", "<=", "MATCHES", "NOT MATCHES", "CONTAINS", "NOT CONTAINS"}
        _validate_allowed_value(self.operator, allowed_operators, "operator")

        if not isinstance(self.value, (bool, int, float, str, datetime)):
            raise TypeError(f"Invalid value type: {type(self.value).__name__}. Expected int, float, str, or datetime.")

    def to_query(self) -> str:
        return f"{self.attribute} {self.operator} {self._right_query()}"

    def _right_query(self) -> str:
        if isinstance(self.value, datetime):
            return f'"{self.value.astimezone().isoformat()}"'
        value = str(self.value)
        value = value.replace("\\", r"\\").replace('"', r"\"")
        return f'"{value}"'


@dataclass
class _AttributePredicate(_Filter):
    postfix_operator: Literal["EXISTS"]
    attribute: _Attribute

    def to_query(self) -> str:
        return f"{self.attribute} {self.postfix_operator}"


@dataclass
class _AssociativeOperator(_Filter):
    operator: Literal["AND", "OR"]
    filters: Iterable[_Filter]

    def __post_init__(self) -> None:
        allowed_operators = {"AND", "OR"}
        _validate_allowed_value(self.operator, allowed_operators, "operator")
        if not self.filters:
            raise ValueError(f"At least one filter must be provided to combine with {self.operator}.")

    def to_query(self) -> str:
        filter_queries = [f"({child})" for child in self.filters]
        return f" {self.operator} ".join(filter_queries)


@dataclass
class _PrefixOperator(_Filter):
    operator: Literal["NOT"]
    filter_: _Filter

    def to_query(self) -> str:
        return f"{self.operator} ({self.filter_})"



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Iterable,
    Literal,
    Union,
)

from neptune_query.internal import filters as _filters
from neptune_query.internal import pattern as _pattern
from neptune_query.internal.retrieval import attribute_types as types
from neptune_query.internal.util import (
    _validate_allowed_value,
    _validate_list_of_allowed_values,
    _validate_string_or_string_list,
)

__all__ = ["Filter", "AttributeFilter", "Attribute", "KNOWN_TYPES"]


KNOWN_TYPES = frozenset(
    {
        "float",
        "int",
        "string",
        "bool",
        "datetime",
        "float_series",
        "string_set",
        "string_series",
        "file",
        "histogram_series",
        "file_series",
    }
)
AGGREGATION_LAST = ("last",)


class BaseAttributeFilter(ABC):
    def __or__(self, other: "BaseAttributeFilter") -> "BaseAttributeFilter":
        return BaseAttributeFilter.any(self, other)

    def any(*filters: "BaseAttributeFilter") -> "BaseAttributeFilter":
        return _AttributeFilterAlternative(filters=filters)

    @abc.abstractmethod
    def _to_internal(self) -> _filters._BaseAttributeFilter:
        ...


@dataclass
class AttributeFilter(BaseAttributeFilter):
    """Filter to apply to attributes when fetching runs or experiments.

    Use to select specific metrics or other metadata based on various criteria.

    Args:
        name (str|list[str], optional):
            if str given: an extended regular expression to match attribute names.
            if list[str] given: a list of attribute names to match exactly.
        type (str|list[str], optional):
            A list of allowed attribute types (or a single type as str). Defaults to all available types:
                ["float", "int", "string", "bool", "datetime", "float_series", "string_set", "string_series",
                "file", "histogram_series"]
            For reference, see: https://docs.neptune.ai/attribute_types

    # TODO: Update docs post-PY-156
    Example:

    ```
    import neptune_query as npt
    from neptune_query.filters import AttributeFilter


    loss_avg_and_var = AttributeFilter(
        type=["float_series"],
        name="loss$",
        aggregations=["average", "variance"],
    )

    npt.fetch_experiments_table(attributes=loss_avg_and_var)
    ```
    """

    # Stopping formatting here to allow long Literal lines
    # fmt: off
    name: Union[str, list[str], None] = None
    type: Union[
        Literal["bool", "datetime", "file", "float", "int", "string", "string_set", "float_series", "histogram_series", "string_series", "file_series"],  # noqa: E501
        list[
            Literal["bool", "datetime", "file", "float", "int", "string", "string_set", "float_series", "histogram_series", "string_series", "file_series"],  # noqa: E501
        ],
        None,
    ] = None
    # fmt: on

    def __post_init__(self) -> None:
        self.type = self.type or list(KNOWN_TYPES)
        if isinstance(self.type, str):
            self.type = [self.type]
        _validate_string_or_string_list(self.name, "name")
        _validate_list_of_allowed_values(self.type, KNOWN_TYPES, "type")

    def _to_internal(self) -> _filters._AttributeFilter:
        if isinstance(self.name, str):
            return _pattern.build_extended_regex_attribute_filter(
                self.name,
                type_in=self.type,
                aggregations=AGGREGATION_LAST,
            )

        if self.name is None:
            return _filters._AttributeFilter(
                type_in=self.type,
                aggregations=AGGREGATION_LAST,
            )

        if self.name == []:
            raise ValueError(
                "Invalid type for `name` attribute. Expected str, non-empty list of str, or None, but got empty list."
            )

        if isinstance(self.name, list):
            return _filters._AttributeFilter(
                name_eq=self.name,
                type_in=self.type,
                aggregations=AGGREGATION_LAST,
            )

        raise ValueError(
            "Invalid type for `name` attribute. Expected str, non-empty list of str, or None, but got "
            f"{type(self.name)}."
        )


@dataclass
class _AttributeFilterAlternative(BaseAttributeFilter):
    filters: Iterable[BaseAttributeFilter]

    def __or__(self, other: "BaseAttributeFilter") -> "BaseAttributeFilter":
        filters = tuple(self.filters) + (other,)
        return _AttributeFilterAlternative(filters)

    def _to_internal(self) -> _filters._AttributeFilterAlternative:
        return _filters._AttributeFilterAlternative(
            filters=[f._to_internal() for f in self.filters],
        )


@dataclass
class Attribute:
    """Helper for specifying an attribute and its type.

    When fetching experiments or runs, use this class to filter and sort the returned entries.

    Args:
        name (str): An attribute name to match exactly.
        type (Literal["float", "int", "string", "bool", "datetime", "float_series", "string_set"], optional):
            Attribute type. Specify it to resolve ambiguity, in case some of the project's runs contain attributes
            that have the same name but are of a different type.
            For a reference, see: https://docs.neptune.ai/attribute_types

     # TODO: Update docs post-PY-156
    Example:

    Select a metric and pick variance as the aggregation:

    ```
    import neptune_query as npt
    from neptune_query.filters import Attribute, Filter


    val_loss_variance = Attribute(
        name="val/loss",
        aggregation="variance",
    )
    # Construct a filter and pass it to a fetching or listing method
    tiny_val_loss_variance = Filter.lt(val_loss_variance, 0.01)
    npt.fetch_experiments_table(experiments=tiny_val_loss_variance)
    ```
    """

    # fmt: off
    name: str
    type: Union[
        Literal["bool", "datetime", "file", "float", "int", "string", "string_set", "float_series", "histogram_series", "string_series", "file_series"],  # noqa: E501
        None,
    ] = None
    # fmt: on

    def __post_init__(self) -> None:
        _validate_allowed_value(self.type, types.ALL_TYPES, "type")  # type: ignore

    def _to_internal(self) -> _filters._Attribute:
        return _filters._Attribute(
            name=self.name,
            type=self.type,
        )

    def __str__(self) -> str:
        return self._to_internal().to_query()


class Filter:
    """Filter used to specify criteria when fetching experiments or runs.

    Examples of filters:
        - Name or attribute value must match regular expression.
        - Attribute value must pass a condition, like "greater than 0.9".
        - Attribute of a given name must exist or not exist.

    You can negate a filter or join multiple filters with logical operators.

    Methods available for attribute values:
    - `name()`: Name of experiment matches an extended regular expression or a list of names.
    - `eq()`: Attribute value equals
    - `ne()`: Attribute value doesn't equal
    - `gt()`: Attribute value is greater than
    - `ge()`: Attribute value is greater than or equal to
    - `lt()`: Attribute value is less than
    - `le()`: Attribute value is less than or equal to
    - `matches()`: Name of experiment matches an extended regular expression
    - `contains_all()`: Tagset contains all tags, or string contains substrings
    - `contains_none()`: Tagset doesn't contain any of the tags, or string doesn't contain the substrings
    - `exists()`: Attribute exists

    Examples:

    ```
    import neptune_query as npt
    from neptune_query.filters import Filter

    # Fetch metadata from specific experiments
    specific_experiments = Filter.name(["flying-123", "swimming-77"])
    npt.fetch_experiments_table(experiments=specific_experiments)

    # Define various criteria
    owned_by_me = Filter.eq("sys/owner", "vidar")
    loss_filter = Filter.lt("validation/loss", 0.1)
    tag_filter = Filter.contains_none("sys/tags", ["test", "buggy"])
    dataset_check = Filter.exists("dataset_version")

    my_interesting_experiments = owned_by_me & loss_filter & tag_filter & dataset_check
    npt.fetch_experiments_table(experiments=my_interesting_experiments)
    ```
    """

    def __init__(self, internal: _filters._Filter) -> None:
        self._internal = internal

    @staticmethod
    def name(name: Union[str, list[str]]) -> "Filter":
        name_attribute = Attribute(name="sys/name", type="string")
        if isinstance(name, str):
            return Filter.matches(name_attribute, name)
        else:
            return Filter(_filters._Filter.any([_filters._Filter.eq(name_attribute, n) for n in name]))

    @staticmethod
    def eq(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator="==", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def ne(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator="!=", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def gt(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator=">", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def ge(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator=">=", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def lt(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator="<", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def le(attribute: Union[str, Attribute], value: Union[int, float, str, datetime]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributeValuePredicate(operator="<=", attribute=attribute._to_internal(), value=value))

    @staticmethod
    def matches(attribute: Union[str, Attribute], pattern: str) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)

        return Filter(
            _pattern.build_extended_regex_filter(
                attribute=attribute._to_internal(),
                pattern=pattern,
            )
        )

    @staticmethod
    def contains_all(attribute: Union[str, Attribute], values: Union[str, list[str]]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)

        if isinstance(values, str):
            values = [values]

        if values == []:
            raise ValueError(
                "Invalid value for `contains_all` filter. Expected str, or non-empty list of str, but got "
                "an empty list"
            )

        internal_filters = [
            _filters._AttributeValuePredicate(operator="CONTAINS", attribute=attribute._to_internal(), value=value)
            for value in values
        ]

        return Filter(_filters._Filter.all(internal_filters))

    @staticmethod
    def contains_none(attribute: Union[str, Attribute], values: Union[str, list[str]]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)

        if values == []:
            raise ValueError(
                "Invalid value for `contains_none` filter. Expected str, or non-empty list of str, but got "
                "an empty list"
            )

        if isinstance(values, str):
            values = [values]

        internal_filters = [
            _filters._AttributeValuePredicate(operator="NOT CONTAINS", attribute=attribute._to_internal(), value=value)
            for value in values
        ]

        return Filter(_filters._Filter.all(internal_filters))

    @staticmethod
    def exists(attribute: Union[str, Attribute]) -> "Filter":
        if isinstance(attribute, str):
            attribute = Attribute(name=attribute)
        return Filter(_filters._AttributePredicate(postfix_operator="EXISTS", attribute=attribute._to_internal()))

    def __and__(self, other: "Filter") -> "Filter":
        """Logical AND operator to combine two filters."""
        return Filter(_filters._Filter.all([self._to_internal(), other._to_internal()]))

    def __or__(self, other: "Filter") -> "Filter":
        """Logical OR operator to combine two filters."""
        return Filter(_filters._Filter.any([self._to_internal(), other._to_internal()]))

    def __invert__(self) -> "Filter":
        """Logical NOT operator to negate the filter."""
        return Filter(_filters._Filter.negate(self._to_internal()))

    def _to_internal(self) -> _filters._Filter:
        return self._internal

    def __str__(self) -> str:
        return self._to_internal().to_query()



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from collections import defaultdict
from concurrent.futures import Executor
from dataclasses import dataclass
from typing import (
    Generator,
    Generic,
    Iterable,
    Optional,
    TypeVar,
    Union,
)

from neptune_api.client import AuthenticatedClient

from ...exceptions import AttributeTypeInferenceError
from .. import (
    filters,
    identifiers,
)
from ..composition import attribute_components as _components
from ..composition import concurrency
from ..filters import ATTRIBUTE_LITERAL
from ..retrieval import (
    search,
    util,
)
from ..retrieval.attribute_types import (
    FILE_SERIES_AGGREGATIONS,
    FLOAT_SERIES_AGGREGATIONS,
    HISTOGRAM_SERIES_AGGREGATIONS,
    STRING_SERIES_AGGREGATIONS,
)
from ..retrieval.search import ContainerType

T = TypeVar("T")


@dataclass
class AttributeInferenceState:
    original_attribute: filters._Attribute
    attribute: filters._Attribute
    inferred_type: Optional[str] = None
    success_details: Optional[str] = None
    error: Optional[str] = None

    def is_finalized(self) -> bool:
        return self.inferred_type is not None or self.error is not None

    def is_inferred(self) -> bool:
        return self.inferred_type is not None

    def set_success(self, inferred_type: ATTRIBUTE_LITERAL, success_details: Optional[str] = None) -> None:
        self.inferred_type = inferred_type
        self.success_details = success_details
        self.attribute.type = inferred_type

    def set_error(self, error: str) -> None:
        self.error = error


@dataclass
class InferenceState(Generic[T]):
    attributes: list[AttributeInferenceState]
    run_domain_empty: Optional[bool] = None
    result: Optional[T] = None

    @staticmethod
    def empty() -> "InferenceState":
        return InferenceState(attributes=[])

    @staticmethod
    def from_attribute(attribute: filters._Attribute) -> "InferenceState[filters._Attribute]":
        attribute_copy = copy.deepcopy(attribute)

        attribute_states = [
            AttributeInferenceState(
                original_attribute=copy.copy(attribute_copy),
                attribute=attribute_copy,
                inferred_type=attribute_copy.type,
                success_details="Type provided" if attribute_copy.type is not None else None,
            )
        ]
        return InferenceState(attributes=attribute_states, result=attribute_copy)

    @staticmethod
    def from_filter(filter_: filters._Filter) -> "InferenceState[filters._Filter]":
        def _walk_attributes(experiment_filter: filters._Filter) -> Iterable[filters._Attribute]:
            if isinstance(experiment_filter, filters._AttributeValuePredicate):
                yield experiment_filter.attribute
            elif isinstance(experiment_filter, filters._AttributePredicate):
                yield experiment_filter.attribute
            elif isinstance(experiment_filter, filters._AssociativeOperator):
                for child in experiment_filter.filters:
                    yield from _walk_attributes(child)
            elif isinstance(experiment_filter, filters._PrefixOperator):
                yield from _walk_attributes(experiment_filter.filter_)
            else:
                raise RuntimeError(f"Unexpected filter type: {type(experiment_filter)}")

        filter_copy = copy.deepcopy(filter_)

        attributes = _walk_attributes(filter_copy)
        attribute_states = [
            AttributeInferenceState(
                original_attribute=copy.copy(attr),
                attribute=attr,
                inferred_type=attr.type,
                success_details="Type provided" if attr.type is not None else None,
            )
            for attr in attributes
        ]
        return InferenceState(attributes=attribute_states, result=filter_copy)

    def incomplete_attributes(self) -> list[AttributeInferenceState]:
        return [attr_state for attr_state in self.attributes if not attr_state.is_finalized()]

    def is_complete(self) -> bool:
        return all(attr_state.is_finalized() for attr_state in self.attributes)

    def is_run_domain_empty(self) -> bool:
        return self.run_domain_empty is not None and self.run_domain_empty

    def raise_if_incomplete(self) -> None:
        uninferred_attributes = [attr_state for attr_state in self.attributes if not attr_state.is_inferred()]
        if uninferred_attributes:
            attribute_names = [attr_state.original_attribute.name for attr_state in uninferred_attributes]

            details = []
            for attr_state in uninferred_attributes:
                if attr_state.error:
                    details.append(f"{attr_state.original_attribute.name}: {attr_state.error}")
                else:
                    details.append(f"{attr_state.original_attribute.name}: could not find the attribute")

            raise AttributeTypeInferenceError(attribute_names=attribute_names, details=details)

    def get_result_or_raise(self) -> Optional[T]:
        self.raise_if_incomplete()
        return self.result


def infer_attribute_types_in_filter(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[filters._Filter],
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    container_type: search.ContainerType = search.ContainerType.EXPERIMENT,  # TODO: remove the default
) -> InferenceState:
    if filter_ is None:
        return InferenceState.empty()

    state = InferenceState.from_filter(filter_)
    if state.is_complete():
        return state

    _infer_attribute_types_locally(inference_state=state)
    if state.is_complete():
        return state

    _infer_attribute_types_from_api(
        client=client,
        project_identifier=project_identifier,
        filter_=None,
        executor=executor,
        fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        container_type=container_type,
        inference_state=state,
    )
    return state


def infer_attribute_types_in_sort_by(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[filters._Filter],
    sort_by: filters._Attribute,
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    container_type: search.ContainerType = search.ContainerType.EXPERIMENT,  # TODO: remove the default
) -> InferenceState[filters._Attribute]:
    state = InferenceState.from_attribute(sort_by)
    if state.is_complete():
        return state

    _infer_attribute_types_locally(inference_state=state)
    if state.is_complete():
        return state

    _infer_attribute_types_from_api(
        client=client,
        project_identifier=project_identifier,
        filter_=filter_,
        executor=executor,
        fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
        container_type=container_type,
        inference_state=state,
    )
    return state


def _infer_attribute_types_locally(
    inference_state: InferenceState,
) -> None:
    for state in inference_state.incomplete_attributes():
        attribute = state.attribute
        matches = []
        if all(agg in FLOAT_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("float_series")
        if all(agg in STRING_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("string_series")
        if all(agg in FILE_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("file_series")
        if all(agg in HISTOGRAM_SERIES_AGGREGATIONS for agg in attribute.aggregation or []):
            matches.append("histogram_series")
        if len(matches) == 1:
            state.set_success(inferred_type=matches[0], success_details="Inferred from aggregation")


def _infer_attribute_types_from_api(
    client: AuthenticatedClient,
    project_identifier: identifiers.ProjectIdentifier,
    filter_: Optional[filters._Filter],
    executor: Executor,
    fetch_attribute_definitions_executor: Executor,
    container_type: search.ContainerType,
    inference_state: InferenceState,
) -> None:
    attribute_states = inference_state.incomplete_attributes()
    attributes = [state.attribute for state in attribute_states]
    attribute_filter_by_name = filters._AttributeFilter(name_eq=list({attr.name for attr in attributes}))

    output = concurrency.generate_concurrently(
        items=search.fetch_sys_ids(
            client=client,
            project_identifier=project_identifier,
            filter_=filter_,
            container_type=container_type,
        ),
        executor=executor,
        downstream=lambda sys_ids_page: concurrency.fork_concurrently(
            executor=executor,
            downstreams=[
                lambda: _components.fetch_attribute_definitions_split(
                    client=client,
                    project_identifier=project_identifier,
                    attribute_filter=attribute_filter_by_name,
                    executor=executor,
                    fetch_attribute_definitions_executor=fetch_attribute_definitions_executor,
                    sys_ids=sys_ids_page.items,
                    downstream=lambda _, definitions: concurrency.return_value(definitions),
                ),
                lambda: concurrency.return_value(sys_ids_page.items),
            ],
        ),
    )

    results: Generator[
        Union[list[identifiers.SysId], util.Page[identifiers.AttributeDefinition]], None, None
    ] = concurrency.gather_results(output)

    sys_ids: list[identifiers.SysId] = []
    attribute_name_to_definition: dict[str, set[str]] = defaultdict(set)
    for result in results:
        if isinstance(result, util.Page):
            for attr_def in result.items:
                attribute_name_to_definition[attr_def.name].add(attr_def.type)
        elif isinstance(result, list):
            sys_ids.extend(result)

    for state in attribute_states:
        attribute = state.attribute
        if attribute.name in attribute_name_to_definition:
            types = attribute_name_to_definition[attribute.name]
            if len(types) == 1:
                state.set_success(
                    inferred_type=next(iter(types)),  # type: ignore
                    success_details="Inferred from neptune api",
                )
            elif len(types) > 1:
                container_name = "runs" if container_type == ContainerType.RUN else "experiments"
                state.set_error(
                    error=f"Neptune found the attribute name in multiple {container_name} "
                    f"with conflicting types: {', '.join(types)}"
                )

    inference_state.run_domain_empty = len(sys_ids) == 0



#
# Copyright (c) 2025, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass

from .filters import (
    AGGREGATION_LITERAL,
    ATTRIBUTE_LITERAL,
    _Attribute,
    _AttributeFilter,
    _AttributeNameFilter,
    _Filter,
)

_WS_PATTERN = re.compile(r"[ \t\r\n]")
_OR_PATTERN = re.compile(rf"{_WS_PATTERN.pattern}+\|{_WS_PATTERN.pattern}+")
_AND_PATTERN = re.compile(rf"{_WS_PATTERN.pattern}+&{_WS_PATTERN.pattern}+")
_NOT_PATTERN = re.compile(rf"!{_WS_PATTERN.pattern}*")
_REPLACEMENTS = {
    r"\x20": " ",
    r"\x21": "!",
}


@dataclass
class Conjunction:
    positive_patterns: list[str]
    negated_patterns: list[str]


@dataclass
class Alternative:
    children: list[Conjunction]


def parse_extended_regex(pattern: str) -> Alternative:
    pattern = pattern.strip()

    def transform_pattern(part: str) -> str:
        for old, new in _REPLACEMENTS.items():
            part = part.replace(old, new)
        return part

    def parse_conjunction(conjunction_part: str) -> Conjunction:
        """Parse a conjunction part into positive and negated patterns."""
        positive_patterns = []
        negated_patterns = []

        for and_part in _AND_PATTERN.split(conjunction_part):
            if _NOT_PATTERN.match(and_part):
                negated_pattern = _NOT_PATTERN.sub("", and_part, count=1)
                negated_pattern = transform_pattern(negated_pattern)
                negated_patterns.append(negated_pattern)
            else:
                and_part = transform_pattern(and_part)
                positive_patterns.append(and_part)

        return Conjunction(positive_patterns, negated_patterns)

    def parse_alternatives(alternative_part: str) -> list[Conjunction]:
        """Split the pattern by OR and parse each part as a conjunction."""
        return [parse_conjunction(part) for part in _OR_PATTERN.split(alternative_part)]

    # Parse the pattern into alternatives
    conjunctions = parse_alternatives(pattern)
    return Alternative(conjunctions)


def build_extended_regex_filter(attribute: _Attribute, pattern: str) -> _Filter:
    parsed = parse_extended_regex(pattern)

    return _Filter.any(
        [
            _Filter.all(
                [_Filter.matches_all(attribute, pattern) for pattern in conj.positive_patterns]
                + [_Filter.matches_none(attribute, pattern) for pattern in conj.negated_patterns]
            )
            for conj in parsed.children
        ]
    )


def build_extended_regex_attribute_filter(
    pattern: str,
    type_in: list[ATTRIBUTE_LITERAL],
    aggregations: list[AGGREGATION_LITERAL],
) -> _AttributeFilter:
    parsed = parse_extended_regex(pattern)

    return _AttributeFilter(
        must_match_any=[
            _AttributeNameFilter(
                must_match_regexes=conj.positive_patterns if conj.positive_patterns else None,
                must_not_match_regexes=conj.negated_patterns if conj.negated_patterns else None,
            )
            for conj in parsed.children
        ],
        aggregations=aggregations,
        type_in=type_in,
    )
